package com.loafle.overflow.module.sensor.service;

import com.loafle.overflow.commons.model.PageParams;
import com.loafle.overflow.commons.utils.PageUtil;
import com.loafle.overflow.commons.utils.StringConvertor;
import com.loafle.overflow.crawler.config.*;
import com.loafle.overflow.module.domain.model.Domain;
import com.loafle.overflow.module.infra.model.Infra;
import com.loafle.overflow.module.infra.model.InfraHost;
import com.loafle.overflow.module.infra.service.InfraService;
import com.loafle.overflow.module.meta.model.MetaCrawler;
import com.loafle.overflow.module.meta.model.MetaSensorItemKey;
import com.loafle.overflow.module.meta.model.MetaSensorStatus;
import com.loafle.overflow.module.meta.service.MetaSensorItemKeyService;
import com.loafle.overflow.module.probe.model.Probe;
import com.loafle.overflow.module.probe.service.ProbeService;
import com.loafle.overflow.module.sensor.dao.SensorDAO;
import com.loafle.overflow.module.sensor.model.Sensor;
import com.loafle.overflow.module.sensor.model.SensorItem;
import com.loafle.overflow.module.target.model.Target;
import org.codehaus.jackson.map.ObjectMapper;
import org.codehaus.jackson.type.TypeReference;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.domain.Page;
import org.springframework.stereotype.Service;

import javax.transaction.Transactional;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Created by insanity on 17. 6. 28.
 */
@Service("SensorService")
public class SensorService  {

    @Autowired
    SensorDAO sensorDAO;

    @Autowired
    private ProbeService probeService;

    @Autowired
    private InfraService infraService;

    @Autowired
    private SensorItemService sensorItemService;

    @Autowired
    private MetaSensorItemKeyService metaSensorItemKeyService;

    private Map<Short, Map<Integer, MetaSensorItemKey>> mappingMap = null;

    public Sensor regist(Sensor sensor) {
        return this.sensorDAO.save(sensor);
    }

//    public List<Sensor> readAllByTarget(Target target, PageParams pageParams) {
//        return this.sensorDAO.findAllByTarget(target, PageUtil.getPageRequest(pageParams));
//    }

    public Page<Sensor> readAllByDomain(Domain domain, PageParams pageParams) {

        List<Probe> probeList = this.probeService.readAllByDomain(domain);

        List<Target> targetList = this.infraService.readAllTargetByProbeList(probeList);

        return this.sensorDAO.findAllByTargetList(targetList, PageUtil.getPageRequest(pageParams));
    }

    public Page<Sensor> readAllByInfra(long infraId, PageParams pageParams) {
        Infra dbInfra = this.infraService.read(infraId);

        if(dbInfra == null || dbInfra.getTarget() == null) return null;

        return this.sensorDAO.findAllByTarget(dbInfra.getTarget(), PageUtil.getPageRequest(pageParams));

    }

    public Sensor read(String id) {
        return this.sensorDAO.findOne(Long.valueOf(id));
    }

    public void remove(Sensor sensor) {
        this.sensorDAO.delete(sensor);
    }

    public Sensor start(Sensor sensor) {
        MetaSensorStatus status = new MetaSensorStatus((short)1);
        sensor.setStatus(status);
        return this.sensorDAO.save(sensor);
    }

    public Sensor stop(Sensor sensor) {
        MetaSensorStatus status = new MetaSensorStatus((short)2);
        sensor.setStatus(status);
        return this.sensorDAO.save(sensor);
    }

    @Transactional
    public Sensor registSensorConfig(Sensor sensor, List<SensorItem> sensorItemList, String etcJson) {

        this.sensorDAO.save(sensor);

        for(SensorItem sensorItem : sensorItemList) {
            sensorItem.setSensor(sensor);
        }

        this.sensorItemService.registAll(sensorItemList);

        return sensor;
    }

    public String generateSensorConfig(Sensor sensor) throws IOException {

        PageParams pageParams = new PageParams();
        pageParams.setPageNo(0);
        pageParams.setCountPerPage(Integer.MAX_VALUE);
        pageParams.setSortCol("id");
        pageParams.setSortDirection("descending");
        Page<SensorItem> dbItemList = this.sensorItemService.readAllBySensor(sensor, pageParams);

        List<SensorItem> sensorItems = dbItemList.getContent();

        if(sensorItems.size() <= 0) {
            return "";
        }
        Sensor dbSensor = sensorItems.get(0).getSensor();

        Infra infra = this.infraService.readByTarget(dbSensor.getTarget());

        // 7 = Infra OS Service
        if(infra.getInfraType().getId() == 7) {
            return this.processService(dbSensor, sensorItems, infra);
        }
        if(infra.getInfraType().getId() == 2) {
            return this.processHost(dbSensor, sensorItems, infra);
        }

        return null;
    }

    private String processHost(Sensor dbSensor, List<SensorItem> sensorItems, Infra infra) throws IOException {

        InfraHost infraHost = (InfraHost)infra;

        Config config = new Config();
        config.setId(String.valueOf(dbSensor.getId()));

        com.loafle.overflow.crawler.config.Target target = this.createTarget(infraHost, dbSensor);

        config.setTarget(target);

        // FIXME: Interval
        Schedule schedule = new Schedule();
        schedule.setInterval("5");
        config.setSchedule(schedule);

        Crawler crawler = this.getCrawler(dbSensor.getCrawler());
        config.setCrawler(crawler);

        Map<Integer, MetaSensorItemKey> keyMap = this.initMappingMap(dbSensor.getCrawler());

        switch (dbSensor.getCrawler().getId()) {
            case 23:
                processWMI(sensorItems, keyMap, dbSensor, config);
            default:
                break;
        }

        ObjectMapper objectMapper = new ObjectMapper();
        return objectMapper.writeValueAsString(config);
    }

    private com.loafle.overflow.crawler.config.Target createTarget(InfraHost infraHost, Sensor dbSensor) {

        com.loafle.overflow.crawler.config.Target target = new com.loafle.overflow.crawler.config.Target();
        Connection connection = new Connection();
        connection.setIp(StringConvertor.intToIp(infraHost.getIp()));


        switch (dbSensor.getCrawler().getId()) {
            case 23:
            {
                connection.setPort("135");
                connection.setPortType("tcp");
                connection.setSsl(false);

                target.setConnection(connection);

                // FIXME: load crawler Auth Info
                Map<String, Object> auth = new HashMap<>();
                auth.put("id", "administrator");
                auth.put("pw", "dbseogns18");

                target.setAuth(auth);

            }
                break;
        }


        return target;
    }

    private String processService(Sensor dbSensor, List<SensorItem> sensorItems, Infra infra) throws IOException {
        com.loafle.overflow.module.infra.model.InfraService infraService = (com.loafle.overflow.module.infra.model.InfraService)infra;

        Config config = new Config();
        config.setId(String.valueOf(dbSensor.getId()));

        com.loafle.overflow.crawler.config.Target target = new com.loafle.overflow.crawler.config.Target();
        Connection connection = new Connection();
        connection.setIp(StringConvertor.intToIp(infraService.getHost().getIp()));
        connection.setPort(String.valueOf(infraService.getPort()));
        connection.setPortType(infraService.getPortType());
        connection.setSsl(infraService.isTlsType());

        target.setConnection(connection);

        Map<String, Object> auth = this.getAuth(dbSensor.getCrawler(), infraService);

        target.setAuth(auth);

        config.setTarget(target);

        // FIXME: Interval
        Schedule schedule = new Schedule();
        schedule.setInterval("5");
        config.setSchedule(schedule);

        Crawler crawler = this.getCrawler(dbSensor.getCrawler());
        config.setCrawler(crawler);

        Map<Integer, MetaSensorItemKey> keyMap = this.initMappingMap(dbSensor.getCrawler());

        switch (dbSensor.getCrawler().getId()) {
            case 11:
                processMysql(sensorItems, keyMap, dbSensor, config);
            case 23:
                processWMI(sensorItems, keyMap, dbSensor, config);
            default:
                break;
        }

        ObjectMapper objectMapper = new ObjectMapper();
        return objectMapper.writeValueAsString(config);
    }

    private void processMysql(List<SensorItem> sensorItems, Map<Integer, MetaSensorItemKey> KeyMap, Sensor dbSensor, Config config) throws IOException {
        Item item = new Item();
        Keys keys = null;
        List<Keys> keysList = new ArrayList<>();
        for(SensorItem sItem : sensorItems) {
            keys = new Keys();
            keys.setMetric(sItem.getItem().getKey());
            keys.setKey(KeyMap.get(sItem.getItem().getId()).getKey());
            keysList.add(keys);
        }
        item.setKeys(keysList);

        QueryInfo queryInfo = new QueryInfo();
        MappingInfo mappingInfo = new MappingInfo();

        this.setQueryAndMapping(dbSensor.getCrawler(), keysList, queryInfo, mappingInfo);

        item.setQueryInfo(queryInfo);
        item.setMappingInfo(mappingInfo);

        List<Item> itemList = new ArrayList<>();
        itemList.add(item);

        config.setItems(itemList);


//        ObjectMapper objectMapper = new ObjectMapper();
//
//        return objectMapper.writeValueAsString(config);
    }
    private void processWMI(List<SensorItem> sensorItems, Map<Integer, MetaSensorItemKey> keyMap, Sensor dbSensor, Config config) throws IOException {

        Map<String, List<MetaSensorItemKey>> metricMap = new HashMap<>();

        MetaSensorItemKey itemKey = null;

        for(SensorItem sItem : sensorItems) {
            itemKey = keyMap.get(sItem.getItem().getId());
            if(metricMap.containsKey(itemKey.getFroms()) == false) {
                metricMap.put(itemKey.getFroms(), new ArrayList<>());
            }
            metricMap.get(itemKey.getFroms()).add(itemKey);
        }

        List<Item> itemList = new ArrayList<>();
        Item item = null;
        QueryInfo queryInfo = null;
        MappingInfo mappingInfo = null;

        StringBuffer stringBuffer = new StringBuffer();

        Map<String, Object> extendMap = new HashMap<>();
        extendMap.put("nameSpace", "root/cimv2");
        extendMap.put("wmicPath", "/home/snoop/temp/wmic");


        List<Keys> keysList = null;
        MetaSensorItemKey tempItemKey = null;
        Keys keys = null;
        for(Map.Entry<String, List<MetaSensorItemKey>> elems : metricMap.entrySet()) {
            keysList =  new ArrayList<>();
            item = new Item();
            stringBuffer.setLength(0);

            stringBuffer.append("SELECT ");

            for(int indexI = 0; indexI < elems.getValue().size(); ++indexI) {

                tempItemKey = elems.getValue().get(indexI);
                stringBuffer.append(tempItemKey.getKey());
                if(indexI + 1 < elems.getValue().size()) {
                    stringBuffer.append(", ");
                }

                keys = new Keys();
                keys.setKey(tempItemKey.getKey());
                keys.setMetric(tempItemKey.getItem().getKey());
                keysList.add(keys);
            }


            List<String> arrayColumns = null;

            String json = tempItemKey.getOption();
            if(json != null && json.length() > 0) {
                HashMap<String,Object> optionMap;
                optionMap = new ObjectMapper().readValue(json, new TypeReference<HashMap<String,Object>>() {});

                Object obj = null;
                obj = optionMap.get("appends");
                if(obj != null) {
                    List<String> appendsList = (List<String>)obj;
                    for(String append : appendsList) {
                        stringBuffer.append(", ");
                        stringBuffer.append(append);
                    }
                }

                stringBuffer.append(" FROM ");
                stringBuffer.append(elems.getKey());

                obj = optionMap.get("where");
                if(obj != null) {
                    String where = (String)obj;
                    stringBuffer.append(" WHERE ");
                    stringBuffer.append(where);
                }

                obj = optionMap.get("arrayColumns");
                if(obj != null) {
                    arrayColumns = (List<String>)obj;
                }

            }
            else {
                stringBuffer.append(" FROM ");
                stringBuffer.append(elems.getKey());
            }

            queryInfo = new QueryInfo();
            queryInfo.setQuery(stringBuffer.toString());
            queryInfo.setExtend(extendMap);

            mappingInfo = new MappingInfo();
            mappingInfo.setParseDirection("col");
            mappingInfo.setArrayColumns(arrayColumns);

            item.setMappingInfo(mappingInfo);
            item.setQueryInfo(queryInfo);
            item.setKeys(keysList);

            itemList.add(item);

        }

        config.setItems(itemList);

    }

    private void setQueryAndMapping(MetaCrawler metaCrawler, List<Keys> keysList, QueryInfo queryInfo, MappingInfo mappingInfo) {

        switch (metaCrawler.getId()) {
            case 11: // mysql
            {
                String query = "show status where ";
//                queryInfo.setQuery("show status where ");

                Keys keys = null;
                for(int indexI = 0 ; indexI < keysList.size(); ++indexI) {
                    keys = keysList.get(indexI);
                    query += "variable_name = '";
                    query += keys.getKey();
                    query += "'";
                    if(indexI + 1 < keysList.size()) {
                        query += " or ";
                    }
                }

                queryInfo.setQuery(query);

                mappingInfo.setParseDirection("row");
                mappingInfo.setValueColumn("Value");

                List<String> keyColumns = new ArrayList<>();
                keyColumns.add("Variable_name");

                mappingInfo.setKeyColumns(keyColumns);
            }
            break;
        }

    }

    private Map<Integer, MetaSensorItemKey> initMappingMap(MetaCrawler metaCrawler) {

        if(this.mappingMap == null) {
            this.mappingMap = new HashMap<>();
        }

        Map<Integer, MetaSensorItemKey> resultMap = this.mappingMap.get(metaCrawler.getId());

        if(resultMap != null) {
            return resultMap;
        }

        resultMap = this.metaSensorItemKeyService.readAllByCrawler(metaCrawler);
        this.mappingMap.put(metaCrawler.getId(), resultMap);

        return resultMap;
    }

    private  Map<String, Object> getAuth(MetaCrawler metaCrawler, com.loafle.overflow.module.infra.model.InfraService infraService) {

        Map<String, Object> auth = new HashMap<>();

        switch (metaCrawler.getId()) {
            case 11: // mysql
            {
                auth.put("url", "jdbc:mysql://"+StringConvertor.intToIp(infraService.getHost().getIp())+":"+String.valueOf(infraService.getPort()));
                auth.put("id", "docker"); // FIXME: Auth Info
                auth.put("pw", "qwer5795QWER"); // FIXME: Auth Info
            }
            break;
        }

        return auth;
    }

    private Crawler getCrawler(MetaCrawler metaCrawler) {

        Crawler crawler = new Crawler();
        crawler.setName(metaCrawler.getName());

        String container = "";

        switch (metaCrawler.getId()) {
            case 1:
                break;
            case 2:
            case 9:
            case 10:
            case 11:
            case 13:
            case 15:
                container = "java_proxy";
                break;
            default:
                container = "java_proxy";
                break;
        }


        crawler.setContainer(container);

        return crawler;
    }

}
